/*
 * general.c
 */

#include "general.h"

/* Maak een eigen structuur voor een graaf */

mygraph emptygraph(int n) {
    mygraph h = {n, 0,
        {0},
        {
            {0}},
        {0}};
    return h;
}

void print_sparse_graph_nauty(sparsegraph sparse_graph) {
    int i, j;
    fprintf(stderr, "Printing sparse graph nauty:\n");
    for (i = 0; i < sparse_graph.nv; i++) {
        fprintf(stderr, "%d :", i);
        for (j = 0; j < sparse_graph.d[i]; j++) {
            //fprintf(stderr, " %d", sparse_graph.e[i * REG + j]);
            fprintf(stderr, " %d", sparse_graph.e[sparse_graph.v[i] + j]);
        }
        fprintf(stderr, "\n");
    }
    fprintf(stderr, "Number of directed edges: %lu\n", (unsigned long) sparse_graph.nde);
}

void output_graph_g6(mygraph * h) {
    if (!converted_to_sg) {
        mygraph_to_sg(&sg, h);
    }
    //converted_to_sg=1;

    //print_sparse_graph_nauty(sg);

    //printf("%s", sgtog6(&sg));
    //sgtog6(&sg);
    writeg6_sg(stdout, &sg);

    number_of_graphs_written++;
}

void print_mygraph(mygraph h) {
    int n = h.n;
    int i, j, d;
    int nbrs[n];
    for (i = 0; i < n; i++) {
        for (j = 0; j < n; j++) {
            nbrs[j] = 0;
        }
        d = h.deg[i];
        for (j = 0; j < d; j++) {
            nbrs[h.neighbours[i][j]] = 1;
        }
        for (j = 0; j < n; j++) {
            printf("%d ", nbrs[j]);
        }
        printf("\n");
    }
    printf("\n");
}

void print_mygraph_adjlist(mygraph h) {
    int n = h.n;
    int i, j;
    //int nbrs[n];
    for (i = 0; i < n; i++) {
        fprintf(stderr, "%d: ", i);
        for (j = 0; j < h.deg[i]; j++)
            fprintf(stderr, "%d ", h.neighbours[i][j]);
        fprintf(stderr, "\n");
    }
}

void add_edge(mygraph * h, int i, int j) {
    (h -> neighbours)[i][(h->deg)[i]] = j;
    (h->deg)[i]++;
    (h -> neighbours)[j][(h->deg[j])] = i;
    (h->deg)[j]++;
    (h -> e)++;
    h->edges[i] |= BIT(j);
    h->edges[j] |= BIT(i);
}

void remove_edge(mygraph * h, int i, int j) {
    remove_nbr((h -> neighbours[i]), h->deg[i], j);
    remove_nbr((h -> neighbours[j]), h->deg[j], i);
    (h->deg)[i]--;
    (h->deg)[j]--;
    (h -> e)--;
    h->edges[i] &= ~BIT(j);
    h->edges[j] &= ~BIT(i);

}

void remove_last_edge(mygraph * h, int i, int j) {
    (h->deg)[i]--;
    (h->deg)[j]--;
    (h -> e)--;
    h->edges[i] &= ~BIT(j);
    h->edges[j] &= ~BIT(i);
}

void remove_nbr(int * l, int length, int w) {
    int i = 0;
    while (l[i] != w) {
        i++;
    }
    while (i < length) {
        l[i] = l[i + 1];
        i++;
    }
}

/* Zet een graaf in nauty om naar een graaf in mijn datastructuur. De lijsten van buren zijn voor elke top al automatisch gesorteerd. */


mygraph graph_to_mygraph(graph * g, int n, int m) {
    mygraph h = emptygraph(n);
    int i, j;
    for (i = 0; i < n; i++) {
        for (j = (i + 1); j < n; j++) {
            if (ISELEMENT(g, 64 * m * i + j)) {
                add_edge(&h, i, j);
            }
        }
    }
    return h;
}

/* Zet een graaf om naar een nauty-graaf */

void mygraph_to_graph(graph * g, mygraph * h) {
    int n = h -> n;
    EMPTYGRAPH(g, m, n);
    int i, j, d;
    for (i = 0; i < n; i++) {
        d = h->deg[i];
        for (j = 0; j < d; j++) {
            ADDONEARC(g, i, (h->neighbours)[i][j], m);
        }
    }
}

//sparsegraph * graph_to_sparsegraph(mygraph * h){
//	int n = h->n;
//	int m = SETWORDSNEEDED(n);
//	graph g[n*m];
//	mygraph_to_graph(g,h);
//	sparsegraph * sg;
//	nauty_to_sg(g, NULL, m, n);
//	return sg;
//}

void mygraph_to_sg(sparsegraph * sparse, mygraph * h) {

    int i, j, k;
    int n = h->n;

    sparse->nde = 0;
    sparse->nv = n;

    j = 0;
    for (i = 0; i < n; i++) {
        sparse->v[i] = j;
        for (k = 0; k < h->deg[i]; k++) {
            sparse->e[j] = h->neighbours[i][k];
            j++;
            sparse->nde++;
        }
        sparse->d[i] = j - sparse->v[i];
    }
}

/* Geeft een cykelgraaf terug */

mygraph cycle_graph(int n) {
    mygraph h = emptygraph(n);
    int i;
    for (i = 0; i < n - 1; i++) {
        add_edge(&h, i, i + 1);
    }
    add_edge(&h, 0, n - 1);
    return h;
}

void determine_edge_numbers(int n) {
    int i, j;
    for (i = 0; i < n; i++) {
        for (j = i + 1; j < n; j++) {
            int k = j + (n - 1) * i - ((i + 1) * i) / 2 - 1;
            edge_number[i][j] = k;
            edge_number[j][i] = k;
            number_to_edge[k][0] = i;
            number_to_edge[k][1] = j;
        }
    }
}

/* Zet een 64-bit long om naar een array van 64 ints */

void long64_to_bin_array(unsigned long long a, int *b, int length) {
    int i;
    for (i = 0; i < length; i++) {
        *(b + length - 1 - i) = (a >> i) & 1;
    }
}

/* Print een array af */

void printArray(int * arr, int length) {
    int i;
    for (i = 0; i < length; i++) {
        printf("%d ", *(arr + i));
    }
    printf("\n");
}

/* Print adjacentiematrix van een graph af
 */

/*
void printMatrix(graph * g, int n, int m, int length) {
    int i;
    int * arr[length];
    for (i = 0; i < (n * m); i++) {
        long64_to_bin_array(*(g + i), arr, length);
        printArray(arr, n);
    }
    printf("\n\n");
}
*/

/*
void matrix_to_file(char * file, graph * g, int n, int m, int length) {
    int i;
    int * arr[length];
    for (i = 0; i < (n * m); i++) {
        long64_to_bin_array(*(g + i), arr, length);
        printArray(arr, n);
    }
    printf("\n\n");
}
*/

/* Bepaal de graad van een top
 */

int degree(graph * g, int v, int m) {
    int deg = 0;
    int i;
    for (i = 0; i < m; i++) {
        //deg += POPCOUNT(g[v+i]);
        deg += POPC(g[v + i]);
    }
    return deg;
}

int contains(int * arr, int value, int length) {
    int i;
    for (i = 0; i < length; i++) {
        if (*(arr + i) == value) {
            return 1;
        }
    }
    return 0;
}

void print_list(mygraph * h) {
    int n = h->n;
    int i;
    for (i = 0; i < n; i++) {
        printArray(h->neighbours[i], h->deg[i]);
    }
}

int build_n_min_one_cycle(mygraph * h, unsigned long long cycle, int * cycle2, int start, int v, int * degrees, int total, int max, int current_length) {
    cycle2[current_length - 1] = v;
    if (total > max) {
        return total;
    }
    int n = h -> n;
    if (current_length == n) {
        if (adjacent(h, start, v)) {
            memcpy(hamiltonian_cycles + number_of_hamiltonian_cycles, cycle2, sizeof (int)*n);
            number_of_hamiltonian_cycles++;
            return (total + 1);
        }
        return total;
    }
    int d = h -> deg[v];
    int w, i;
    int possible_nbrs[d];
    int length = 0;

    for (i = 0; i < d; i++) {
        w = h -> neighbours[v][i];
        if (!(cycle & BIT(w))) {
            possible_nbrs[length] = w;
            length++;
        }
    }
    for (i = 0; i < length; i++) {
        cycle = (cycle | BIT(possible_nbrs[i]));
        total = build_n_min_one_cycle(h, cycle, cycle2, start, possible_nbrs[i], degrees, total, max, current_length + 1);
        if (total > max) {
            return total;
        }
        cycle = cycle & ~BIT(possible_nbrs[i]);
    }
    return total;
}

int build_ham_cycle(mygraph * h, unsigned long long cycle, int * cycle2, int start, int v, int * degrees, int total, int max, int current_length) {
    cycle2[current_length - 1] = v;
    if (total > max) {
        return total;
    }
    int n = h -> n;
    if (current_length == n) {
        if (adjacent(h, start, v)) {
            memcpy(hamiltonian_cycles + number_of_hamiltonian_cycles, cycle2, sizeof (int)*n);
            number_of_hamiltonian_cycles++;
            return (total + 1);
        }
        return total;
    }
    int d = h -> deg[v];
    int deg, w, i;
    int first_w;
    int possibility_check = 0;

    int possible_nbrs[d];
    int length = 0;

    for (i = 0; i < d; i++) {
        w = h -> neighbours[v][i];
        if (!(cycle & BIT(w))) {
            deg = degrees[w];
            if (deg == 0 || (deg == 1 && !adjacent(h, start, w))) {
                for (i = 0; i < length; i++) {
                    degrees[possible_nbrs[i]]++;
                }
                return total;
            }
            if ((deg == 1 && adjacent(h, start, w) && current_length != (n - 1))) {
                for (i = 0; i < length; i++) {
                    degrees[possible_nbrs[i]]++;
                }
                return total;
            }
            if ((deg == 2 && !adjacent(h, start, w)) || (deg == 1 && adjacent(h, start, w) && current_length == n - 1)) {
                possibility_check++;
                if (possibility_check == 2) {
                    for (i = 0; i < length; i++) {
                        degrees[possible_nbrs[i]]++;
                    }
                    return total;
                }
                first_w = w;
            }
            if (deg != 0) {
                possible_nbrs[length] = w;
                length++;
            }
            degrees[w]--;
        }
    }
    if (possibility_check == 1) {
        cycle = (cycle | BIT(first_w));
        total = build_ham_cycle(h, cycle, cycle2, start, first_w, degrees, total, max, current_length + 1);
        cycle = cycle & ~BIT(first_w);
    } else {
        for (i = 0; i < length; i++) {
            cycle = (cycle | BIT(possible_nbrs[i]));
            total = build_ham_cycle(h, cycle, cycle2, start, possible_nbrs[i], degrees, total, max, current_length + 1);
            if (total > max) {
                return total;
            }
            cycle = cycle & ~BIT(possible_nbrs[i]);
        }
    }
    for (i = 0; i < length; i++) {
        degrees[possible_nbrs[i]]++;
    }
    return total;
}

int hamiltonian_cycle(mygraph * h, int max) {
    number_of_hamiltonian_cycles = 0;
    int total = 0;
    int n = h -> n;
    int i, deg;
    int min_v = 0;
    int min_deg = n;
    for (i = 0; i < n; i++) {
        deg = h -> deg[i];
        if (deg == 0 || deg == 1) {
            return 0;
        } else {
            if (deg < min_deg) {
                min_v = i;
                min_deg = deg;
            }
        }
    }

    int j, v, w;
    int * d = h -> neighbours[min_v];

    /* Houdt een lijst bij met het aantal mogelijke buren dat een top nog heeft om naartoe te gaan tijdens het bouwen van de cykel.
     * We verlagen het aantal mogelijkheden al met 1 voor alle buren van min_v (geen enkele top kan nog naar daar gaan tijdens het proces) */

    int degrees[n];
    memcpy(degrees, h -> deg, n * sizeof (int));

    for (j = 0; j < (h->deg[min_v]); j++) {
        degrees[h->neighbours[min_v][j]]--;
    }

    /* We overlopen alle paren van buren van min_v. Voor elk paar v-w starten we de cykel als v - min_v - w -...
     * We verlagen de mogelijkheden voor alle buren van v.
     * We starten de recursie vanaf top w. We proberen stapsgewijs een cykel te maken, als de cykel lengte n heeft kijken we of de laatste top adjacent is met v.
     * Onderweg snoeien we een beetje door naar het aantal mogelijke toppen te kijken dat we nog kunnen kiezen.
     */

    for (i = 0; i < min_deg; i++) {
        v = d[i];
        for (j = 0; j < (h->deg[v]); j++) {
            degrees[h->neighbours[v][j]]--;
        }
        for (j = (i + 1); j < min_deg; j++) {
            w = d[j];
            int cycle2[n];
            cycle2[0] = v;
            cycle2[1] = min_v;
            cycle2[2] = w;
            unsigned long long cycle = BIT(v) | BIT(min_v) | BIT(w);
            total = build_ham_cycle(h, cycle, cycle2, v, w, degrees, total, max, 3);
            if (total > max) {
                return total;
            }
        }
        for (j = 0; j < (h->deg[v]); j++) {
            degrees[h->neighbours[v][j]]++;
        }
    }
    return total;
}

void print_help(char * argv0) {
    fprintf(stderr, "Usage: %s <number_of_vertices> [options]\n", argv0);
    fprintf(stderr, "\n");
    fprintf(stderr, "Valid options are:\n");
    fprintf(stderr, " H<x>: Generate graphs with exactly <x> hamiltonian cycles (default 1)\n");
    fprintf(stderr, " o   : Output hamiltonian graphs with at most the specified number of hamiltonian cycles (instead of exactly that number of cycles, only if the number is > 1)\n");
    fprintf(stderr, " D<x>: Only output graphs with maximum degree at most <x>\n");
    fprintf(stderr, " M<x>: Only output graphs with at most <x> nodes of maximum degree (maximum degree must be specified)\n");
    fprintf(stderr, " T   : Only output triangle-free graphs\n");
    fprintf(stderr, " F   : Only output square-free graphs\n");
    fprintf(stderr, " P   : Only output planar graphs\n");
    fprintf(stderr, " C   : Only output graphs without n-1 cycles\n");
    fprintf(stderr, " S <rest> <modulo>: Splits the generation in <modulo> parts where part <rest> will be executed.\n");
}


