/*
 * general.h
 */

#ifndef GENERAL_H_
#define GENERAL_H_

//#define MAXN 64

#include <stdio.h>
#include <stdlib.h>
//#include <naututil.h>
#include <string.h>
#include <math.h>
#include "nausparse.h"
#include <sys/times.h>
#include "planarity.h" /* which includes nauty.h */
//#include <gtools.c>


//#define N 64
#define N MAXN
//#define BIT(i) (1ULL << (i))


#if WORDSIZE == 64
#define BIT(i) (1ULL << (i))
#else //i.e. WORDSIZE == 32
#define BIT(i) (1U << (i))
#endif

//Uncomment to use builtin popcount (else uses nauty popcount)
//Need to compile with gcc -mpopcnt or -msse4.2 to enable __builtin_popcount!
//USE_BUILTIN_POPCOUNT is significantly faster for 64bit
#define USE_BUILTIN_POPCOUNT

#ifdef USE_BUILTIN_POPCOUNT

#if WORDSIZE == 64
#define POPC(x) (__builtin_popcountll(x)) 
#else //i.e. WORDSIZE == 32
#define POPC(x) (__builtin_popcount(x)) 
#endif

#define POPCLL(x) (__builtin_popcountll(x)) 

#else
#define POPC(x) POPCOUNT(x) /* needs nauty.h */
#endif

#define C 1/(log(3)/log(2) - 1)
#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))
#define adjacent(h,i,j) ((h->edges[(i)] & BIT((j))))
#define time_factor sysconf(_SC_CLK_TCK)
#define max_edges(n) ((n)*(n))/4+1

graph g[N];
int permutations[MAXN][N];
int number_of_permutations;
int edge_number[N][N];
int number_to_edge[N*(N - 1) / 2][2];
int m;
unsigned long long int number_of_nauty_calls;
int number_of_edges_before_nauty_call;
int number_of_non_canonical_edges;
int maxdeg;
int max_number_of_cycles;
int number_of_vertices;
int u;
int triangle_free;
int square_free;
int planar;
int mindeg_2;
int current_number_of_cycles;
int nauty_called;
int possible_edges_to_add[N*(N - 1) / 2];
int number_of_edges_to_add;
int reducible_edges[N*(N - 1) / 2];
int number_of_reducible_edges;
int hamiltonian_cycles[MAXN][N];
int number_of_hamiltonian_cycles;
int new_cycles;
int max_number_of_nodes_maxdeg;
int current_number_of_nodes_maxdeg;
int current_number_of_nodes_mindeg;
int converted_to_sg;
unsigned long long int number_of_graphs_tested;
unsigned long long int number_of_graphs_by_criterium[7];
unsigned long long int num_graphs_tested_for_planarity;
unsigned long long int num_graphs_rejected_for_planarity;
int no_n_minus_one;
int number_of_times_max_edges;
int edges_in_cycle[N];
int parent_of_edge[(N*(N - 1)) / 2];
int tree_size_of_root[(N*(N - 1)) / 2];
unsigned long long bit_cycle[N];
int number_added_to_reducible_edges;
int reducible_edges_in_cycle[N];
int number_of_reducible_edges_in_cycle;
int number_of_cycles_through_edge[N];
int changed_cycle_edges[N];
int number_of_changed_cycle_edges;
int modulo;
int splitlevel_counter;
int mod;
int rest;
int splitlevel;
//unsigned long long int number_of_graphs_with_edges[N*N] = {0};
unsigned long long int number_of_graphs_with_given_num_edges[N*N];
int canonical_edge_found;
int output_exact;
unsigned long long int reject_by_ham_cycle;
int u;
int not_canonical;

signed int criteria[7];
int number_of_criteria;

unsigned long long int number_of_graphs_written;

int lab[MAXN];
int ptn[MAXN];
int orbits[MAXN];
//
//DYNALLSTAT(int,lab,lab_sz);
//DYNALLSTAT(int,ptn,ptn_sz);
//DYNALLSTAT(int,orbits,orbits_sz);

statsblk stats;
sparsegraph sg;
sparsegraph canonsg;

static DEFAULTOPTIONS_SPARSEGRAPH(options);

typedef struct {
    int n;
    int e;
    int deg[N];
    int neighbours[N][N]; //Adjacency list
    unsigned long long edges[N]; //Bitvector of neighbours
} mygraph;


int max_min_degree(int n);
int allowed_to_add(mygraph *, int, int);
void expand(mygraph *);
void generate();
int allowed_to_add_tf(mygraph *, int, int);
void expand_tf(mygraph *, int, int);
void determine_orbits(int * edges, int number_of_edges, int * parent_of_edge, int * tree_size, int n);


mygraph emptygraph(int);
void print_mygraph(mygraph);
void print_mygraph_adjlist(mygraph);
void print_sparse_graph_nauty(sparsegraph);
void add_edge(mygraph *, int, int);
void remove_nbr(int * l, int length, int w);
void remove_edge(mygraph *, int, int);
void remove_last_edge(mygraph *, int, int);
mygraph graph_to_mygraph(graph *, int, int);
void mygraph_to_graph(graph *, mygraph *);
mygraph cycle_graph(int);
void print_cycle(int, int);
void print_list(mygraph *);


void long64_to_bin_array(unsigned long long, int *, int);
void printArray(int *, int);
void printMatrix(graph *, int, int, int);
int degree(graph *, int, int);
//int contains(int* , int , int);
//void write_hamiltonian_cycles(char *, char *, int, int, int, int);

int build_ham_cycle(mygraph * h, unsigned long long cycle, int * cycle2, int start, int v, int * degrees, int total, int max, int current_length);
int hamiltonian_cycle(mygraph * h, int max);

void mygraph_to_sg(sparsegraph * sparse, mygraph * h);
void writeg6_sg(FILE*,sparsegraph*);

int build_n_min_one_cycle(mygraph * h, unsigned long long cycle, int * cycle2, int start, int v, int * degrees, int total, int max, int current_length);
void output_graph_g6(mygraph * h);

void determine_orbits(int * edges, int number_of_edges, int * parent_of_edge, int * tree_size, int n);
void determine_edge_numbers(int n);

void print_help(char * argv0);

#endif /* GENERAL_H_ */